'' /////////////////////////////////////////////////////////////////////////////
'' TV Driver for Defender Remake
'' (C) 2007 Steve Waddicor
'  Derived from SimpleNTSC_256x192x256_colors_002.spin
'' //////////////////////////////////////////////////////////////////////////////

OBJ
    g      : "sw_df_globals_003"

VAR
  long  cogon, cog

PUB start(tvptr) : okay

'' Start TV driver - starts a cog
'' returns false if no cog available
''
''   tvptr = pointer to TV parameters

  stop
  okay := cogon := (cog := cognew(@entry,tvptr)) > 0


PUB stop

'' Stop TV driver - frees a cog

  if cogon~
    cogstop(cog)

DAT                     org
entry                   jmp     #initialization         'Jump past the constants

' I want to explain some important constants up front, so I'll declare them here.
' Ordinarily they'd be tucked away after the code.  There is also a lot of skipping
' arounf between CON and DAT sections.  I'm declaring them in the order I can best explain them.
'
' I've used some naming conventions.  These prefixes mean:
' NTSC_ :   The value is a constant of the NTSC system. They would probably be the same no matter
'           what resolution or color mode you are using.  You'd probably only change them if you were
'           converting to PAL.  In which case, good luck to you.
' CHOOSE_ : A choice was made for this particular resolution and color depth.  You might change these.
' CALC_     These were calculated from one or more of the choices you made.  Choose again and you may
'           have to change these.

'***************************************************
'* Color Frequency                                 *
'***************************************************
'
' The NTSC output is a signal with various elements that need timing relative to each other.  So we need some
' sort of clock from which to time all the elements.  The frequency of the color carrier is useful.  This is
' 3.579545 Mhz. A cycle of this clock will take 1sec/3.579545MHz = 279.365ns.

CON  NTSC_color_frequency       =     3_579_545
DAT  NTSC_color_freq            long  NTSC_color_frequency

' Additionally this period of 279ns is divided up into 16 by a Phased Locked Loop circuit (PLL), and it is multiples
' this period that the Video Scale Hardware Register (VSCL) is programmed.  This period of 17.460313ns is called a
' "clock".



'***************************************************
'* Horizontal sync                                 *
'***************************************************
'* A hsync takes 10.9us.  That's 10.9us/17.460313ns = 624 clocks (rounded to an integer).
CON  NTSC_hsync_clocks          =               624

' We are going to send each horizontal sync with a single WAITVID.  We'll aproximate the waveform with
' 16 x 4 "color" pixels.  So we need to program the Video Scale register VSCL.  That looks like this:
'
' %xxxx_xxxx_xxxx_________________________ : Not used
' %_______________PPPP_PPPP_______________ : Clocks per pixel (CPP)
' %_________________________FFFF_FFFF_FFFF : Clocks per frame (CPF)
'
' The number of pixels we get is CPF/CPP.  So for this usage CPS is 624/16 = 39.
DAT  NTSC_hsync_VSCL            long  NTSC_hsync_clocks >> 4 << 12 + NTSC_hsync_clocks

' The hsync is made up of 3 different sorts of signal.  We put them into a palette:
' Color0 = Color NTSC_control_signal_palette (yellow hue at zero value) = $8a
' Color1 = Black = $02
' Color2 = Blanking level = $00
' Color4 = don't care
' Note that the palette is programmed in reverse order:
DAT  NTSC_control_signal_palette long  $00_00_02_8a

' Then we construct a hsync out of 16 pixels of those control signals.
' Note that this is shifted out of the VSU to the right, so LSB goes first.
DAT  NTSC_hsync_pixels          long  %%11_0000_1_2222222_11



'***************************************************
'* Blank lines                                     *
'***************************************************
' The rest of the line after a hsync takes 52.6us.  That's 3008 clocks (to the nearest multiple of 16).
CON  NTSC_active_video_clocks   =     3008

' For blank lines, were' going to programme the VSCL so it slowly outputs 16 pixels over the entire
' length of the active line.  It's all going to be one colour so we don't even need to programme the
' clocks per pixel part, just the clocks per frame.
DAT  NTSC_active_video_VSCL     long  NTSC_active_video_clocks



'***************************************************
'* User graphics lines                             *
'***************************************************
' The important lines at last.  To fit 256 pixels in, we're going to make them 10 clocks each.
' That's 2560 clocks for the user graphics width.  You could use 11 clocks each or some other near value.
' It depends whether you want overscan to the left and right.
CON CHOOSE_clocks_per_gfx_pixel1 = 32
CON CHOOSE_clocks_per_gfx_pixel2 = 16

' Because we're going to use 256 color mode, we're only going to output 4 pixels per WAITVID.  So that
' decides the number of clocks per frame.  4.
CON CALC_clocks_per_gfx_frame1   =     CHOOSE_clocks_per_gfx_pixel1*4                            '32
CON CALC_clocks_per_gfx_frame2   =     CHOOSE_clocks_per_gfx_pixel2*4                            '32
CON CALC_top_overscan            = (244-g#SCANLINES)/2+5
CON CALC_bottom_overscan         = 244-g#SCANLINES-CALC_top_overscan

' Program the VSCL as before.
DAT CALC_user_data_VSCL1         long  (CHOOSE_clocks_per_gfx_pixel1 << 12) + CALC_clocks_per_gfx_frame1  '32
DAT CALC_user_data_VSCL2         long  (CHOOSE_clocks_per_gfx_pixel2 << 12) + CALC_clocks_per_gfx_frame2  '32

' So if we're doing 256 pixels, 4 at a time, that's 256/4 = 64 frames.  64 WAITVIDS.
CON CALC_frames_per_gfx_line1    = 22 '(168)
CON CALC_frames_per_gfx_line2    = 44 '(168)

' There is some extra active video that is not used for user graphics.  This is the overscan.
CON CALC_overscan = NTSC_active_video_clocks-(CALC_frames_per_gfx_line2*CALC_clocks_per_gfx_frame2)        '3008-2176=832
' It maybe that your TV doesn't centre the picture.  We can tweak the overscan left and right to make
' it balance.  28 works quite well on my TV.  It may well be different for yours.
CON CHOOSE_horizontal_offset    = 28

' So we can work out how much overscan to do left (backporch) and right (frontporch).
CON CALC_backporch = (CALC_overscan/2)+CHOOSE_horizontal_offset                                          '416
CON CALC_frontporch = CALC_overscan - (CALC_overscan/2)-CHOOSE_horizontal_offset                         '416

DAT

'***************************************************
'* The code                                        *
'***************************************************
' Start of real code

initialization          ' VCFG: setup Video Configuration register and 3-bit tv DAC pins to output
                        movs    VCFG, #%0000_0111       ' VCFG'S = pinmask (pin31: 0000_0111 : pin24)
                        movd    VCFG, #3                ' VCFG'D = pingroup (grp. 3 i.e. pins 24-31)

                        movi    VCFG, #%0_10_111_000    ' baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband
                                                        ' %0_xx_x_x_x_xxx : Not used
                                                        ' %x_10_x_x_x_xxx : Composite video to top nibble, broadcast to bottom nibble
                                                        ' %x_xx_1_x_x_xxx : 4 color mode
                                                        ' %x_xx_x_1_x_xxx : Enable chroma on broadcast
                                                        ' %x_xx_x_x_1_xxx : Enable chroma on baseband
                                                        ' %x_xx_x_x_x_000 : Broadcast Aural FM bits (don't care)

                        or      DIRA, tvport_mask       ' set DAC pins to output

                        '
                        or      DIRA, #1                ' enable debug LED
                        'mov     OUTA, #1                ' light up debug LED

                        ' CTRA: setup Frequency to Drive Video
                        movi    CTRA,#%00001_111        ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                        mov     r1, NTSC_color_freq     ' r1: Color frequency in Hz (3.579_545MHz)
                        rdlong  v_clkfreq, #0           ' copy clk frequency. (80Mhz)
                        mov     r2, v_clkfreq           ' r2: CLKFREQ (80MHz)
                        call    #dividefract            ' perform r3 = 2^32 * r1 / r2
                        mov     v_freq, r3              ' v_freq now contains frqa.       (191)
                        mov     FRQA, r3                ' set frequency for counter
'-----------------------------------------------------------------------------
'                       'Set up parameters
                        '  PAR points to a small block:
                        '    Long 0: Next line to be displayed.   TV Driver -> Caller
                        '    Long 1: Address of scanline buffer.  TV Driver <- Caller

                        mov     r0,PAR
                        movd    :load_dest,#params
                        mov     r1,#26
:load                   add     r0,#4
                        '       ___
:load_dest              rdlong  0-0,r0
                        '       ^^^ iterates through renderer_params array.
                        add     :load_dest,inc_dest
                        djnz    r1,#:load
'-----------------------------------------------------------------------------
frame_loop
                        mov     current_line,#0
                        'NTSC has 244 "visible" lines, but some of that is overscan etc.
                        '  so pick a number of lines for user graphics e.g. 192
                        '  then display (244-192)/2 = 26 lines of overscan before
                        '  and after the user display.
                        mov     line_loop, #CALC_top_overscan
                        'There's always a horizontal sync at the start of each line.  The clever stuff is
                        '  in the constants described earlier.
:vert_back_porch        mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        'This is just a blank line.  I've made it brown just to show it's there
                        '  but you'd probably want it black.  Notice how the whole line (apart from the
                        '  hsyc is output with one WAITVID.

                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid brown_border_in_color0, #0

                        djnz    line_loop, #:vert_back_porch
'-----------------------------------------------------------------------------
                        'Time to do the user graphics lines.  We're having 218 of them.
                        mov     line_loop, #g#SCANLINES         '(239 so far)
                        'hsync
user_graphics_lines     mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        'The overscan before the user graphics.  Green in this case.
                        mov     VSCL,#CALC_backporch
                        waitvid brown_border_in_color0, #0

                        cmp     current_line,#g#DIVIDER+2 wz,wc
                        'And now at long last, the user graphics.  Program the VSCL for
                        '  4 pixels per frame.
              if_b      mov     VSCL, CALC_user_data_VSCL1
              if_b      movs    :load_num_frames,#CALC_frames_per_gfx_line1
              if_ae     mov     VSCL, CALC_user_data_VSCL2
              if_ae     movs    :load_num_frames,#CALC_frames_per_gfx_line2

                        ' The meat of the driver.  Transfer the scanbuffer data to the TV.
                        mov     ptr,tv_scanline_buffer_adr
:load_num_frames        mov     tile_loop,#0-0
:loop                   rdlong  pixels,ptr
                        add     ptr,#4
                        waitvid pixels,#%%3210
                        djnz    tile_loop,#:loop

                        'Even before we do the overscan, we can inform the caller that the
                        '  we are ready for the next line
                        add     current_line,#1
                        cmp     current_line,#g#SCANLINES  wz,wc
              if_e      mov     current_line,#0

                        'We've finished drawing a user graphics line, whether it was color
                        '  bar or flag.  Now do the overscan bit.  This time in blue.
end_of_flag_line        mov    VSCL,#CALC_frontporch
                        waitvid brown_border_in_color0 , border

                        wrlong  current_line,tv_scanline_request_adr
                        'loop
                        djnz    line_loop, #user_graphics_lines
'-----------------------------------------------------------------------------
                        'We've finished the visible part of the screen so tell the app
                        'that there's a vsync
                        mov     r0,PAR
                        'add     r0,#22*4
                        wrlong  vsync_flag,r0
'-----------------------------------------------------------------------------
                        'Overscan at the bottom of screen.  Same as top of screen, but
                        '  this time in magenta.
                        mov     line_loop, #CALC_bottom_overscan
                        'hsync
vert_front_porch        mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid brown_border_in_color0, #0
                        djnz    line_loop, #vert_front_porch
'-----------------------------------------------------------------------------
                        'This is the vertical sync.  It consists of 3 sections of 6 lines each.
                        ' first do a single line, then do keyboard stuff.
                        mov     line_loop, #6           '(250)
:vsync_higha            mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_2
                        djnz    line_loop, #:vsync_higha
'-----------------------------------------------------------------------------
                        mov     line_loop, #6           '(256)
:vsync_low              mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_low_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_low_2
                        djnz    line_loop, #:vsync_low
'-----------------------------------------------------------------------------
                        mov     line_loop, #6           '(250)
:vsync_highb            mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_2
                        djnz    line_loop, #:vsync_highb
'-----------------------------------------------------------------------------
                        'mov     OUTA, #0

                        jmp     #frame_loop



'-----------------------------------------------------------------------------
                        'Keyboard initialisation
keyboard_init           movd    :par,#kb_pingroup         'load input parameters _pingroup/_locks/_auto
                        mov     kb_x,par
                        add     kb_x,#11*4
                        mov     kb_y,#3
:par                    rdlong  0,kb_x
                        add     :par,kb_dlsb
                        add     kb_x,#4
                        djnz    kb_y,#:par
                        mov     kb_x,kb_pingroup             'get mask for 1st pin
                        andn    kb_x,#1
                        test    kb_pingroup,#1    wz
                        mov     kb_dwmask,#1
                        shl     kb_dwmask,kb_x
                        mov     kb_drmask,kb_dwmask
        if_nz           shl     kb_drmask,#1
                        mov     kb_cwmask,kb_drmask
                        shl     kb_cwmask,#1
                        mov     kb_crmask,kb_cwmask
        if_nz           shl     kb_crmask,#1

        if_nz           or      dira,kb_dwmask             'set directions
        if_nz           or      dira,kb_cwmask

                        mov     kb_head,#0                'reset output parameter _head

                        mov     kb_state,#reset            'initial state
next_state              jmp     kb_state
'------------------------------------------------------------------------------------------------------
'
'
' Reset keyboard
'
reset                   test    kb_pingroup,#1    wz       'z = proto/hybrid, nz=hydra
        if_nz           andn    outa,kb_dwmask             'let PS/2 lines float (pulled up)
        if_nz           andn    outa,kb_cwmask
        if_z            andn    dira,kb_drmask
        if_z            andn    dira,kb_crmask

                        movd    :par,#kb_present          'reset output parameters _present/_states[8]
                        mov     kb_x,#1+8
:par                    mov     0,#0
                        add     :par,kb_dlsb
                        djnz    kb_x,#:par

                        mov     kb_stat,#8                 'set reset flag
                        'fall thru to...                        
'------------------------------------------------------------------------------------------------------
'
'
' Update parameters
'
update                  movd    :par,#kb_head             'update output parameters _head/_present/_states[8]
                        mov     kb_x,par
                        add     kb_x,#1*4
                        mov     kb_y,#1+1+8
:par                    wrlong  0,kb_x
                        add     :par,kb_dlsb
                        add     kb_x,#4
                        djnz    kb_y,#:par

                        test    kb_stat,#8         wc      'if reset flag, transmit reset command
        if_c            mov     kb_data,#$FF
        if_c            mov     kb_state_after_transmit,#newcode
        if_c            mov     kb_state,#transmit
        if_c            jmp     #next_state
                        'fall thru to...        
'------------------------------------------------------------------------------------------------------
'
'
' Get scancode
'
newcode                 mov     kb_stat,#0                 'reset state
                        'fall thru to...        
'------------------------------------------------------------------------------------------------------
:same                   mov     kb_state_after_recieve,#:same_recieved
                        mov     kb_state,#receive
                        jmp     #next_state

:same_recieved          cmp     kb_data,#$83+1     wc      'scancode?

        if_nc           cmp     kb_data,#$AA       wz      'powerup/reset?
        if_nc_and_z     mov     kb_state,#configure
        if_nc_and_z     jmp     #next_state

        if_nc           cmp     kb_data,#$E0       wz      'extended?
        if_nc_and_z     or      kb_stat,#1
        if_nc_and_z     mov     kb_state,#:same
        if_nc_and_z     jmp     #next_state

        if_nc           cmp     kb_data,#$F0       wz      'released?
        if_nc_and_z     or      kb_stat,#2
        if_nc_and_z     mov     kb_state,#:same
        if_nc_and_z     jmp     #next_state

        if_nc           mov     kb_state,#newcode          'unknown, ignore
        if_nc           jmp     #next_state
'
'
' Translate scancode and enter into buffer
'
'                        test    kb_stat,#1         wc      'lookup code with extended flag
'                        rcl     kb_data,#1
'                        call    #look

'                        cmp     kb_data,#0         wz      'if unknown, ignore
'        if_z            mov     kb_state,#newcode
'        if_z            jmp     #next_state

'                        mov     kb_t,kb_states+6             'remember lock keys in _states

'                        mov     kb_x,kb_data                  'set/clear key bit in _states
'                        shr     kb_x,#5
'                        add     kb_x,#kb_states
'                        movd    :reg,kb_x
'                        mov     kb_y,#1
'                        shl     kb_y,kb_data
'                        test    kb_stat,#2         wc
':reg                    muxnc   0,kb_y

'        if_nc           cmpsub  kb_data,#$F0       wc      'if released or shift/ctrl/alt/win, done
'        if_c            mov     kb_state,#update
'        if_c            jmp     #next_state

'                        mov     kb_y,kb_states+7             'get shift/ctrl/alt/win bit pairs
'                        shr     kb_y,#16

'                        cmpsub  kb_data,#$E0       wc      'translate keypad, considering numlock
'        if_c            test    kb_locks,#%100    wz
'        if_c_and_z      add     kb_data,#@keypad1-@table
'        if_c_and_nz     add     kb_data,#@keypad2-@table
'        if_c            call    #look
'        if_c            mov     kb_state,#:flags
'        if_c            jmp     #next_state

'                        cmpsub  kb_data,#$DD       wc      'handle scrlock/capslock/numlock
'        if_c            mov     kb_x,#%001_000
'        if_c            shl     kb_x,kb_data
'        if_c            andn    kb_x,kb_locks
'        if_c            shr     kb_x,#3
'        if_c            shr     kb_t,#29                   'ignore auto-repeat
'        if_c            andn    kb_x,kb_t             wz
'        if_c            xor     kb_locks,kb_x
'        if_c            add     kb_data,#$DD
'        if_c_and_nz     or      kb_stat,#4                 'if change, set configure flag to update leds

'                        test    kb_y,#%11          wz      'get shift into nz

'        if_nz           cmp     kb_data,#$60+1     wc      'check shift1
'        if_nz_and_c     cmpsub  kb_data,#$5B       wc
'        if_nz_and_c     add     kb_data,#@shift1-@table
'        if_nz_and_c     call    #look
'        if_nz_and_c     andn    kb_y,#%11

'        if_nz           cmp     kb_data,#$3D+1     wc      'check shift2
'        if_nz_and_c     cmpsub  kb_data,#$27       wc
'        if_nz_and_c     add     kb_data,#@shift2-@table
'        if_nz_and_c     call    #look
'        if_nz_and_c     andn    kb_y,#%11
{
                        test    kb_locks,#%010    wc      'check shift-alpha, considering capslock
                        muxnc   :shift,#$20
                        test    kb_locks,#$40     wc
        if_nz_and_nc    xor     :shift,#$20
                        cmp     kb_data,#"z"+1     wc
        if_c            cmpsub  kb_data,#"a"       wc
:shift  if_c            add     kb_data,#"A"
        if_c            andn    kb_y,#%11

:flags                  ror     kb_data,#8                 'add shift/ctrl/alt/win flags
                        mov     kb_x,#4                    '+$100 if shift
:loop                   test    kb_y,#%11          wz      '+$200 if ctrl
                        shr     kb_y,#2                    '+$400 if alt
        if_nz           or      kb_data,#1                 '+$800 if win
                        ror     kb_data,#1
                        djnz    kb_x,#:loop
                        rol     kb_data,#12

                        mov     kb_x,par                   'check for a "break key" key code
                        add     kb_x,#19*4                 '@par_break - @par_head
                        rdword  kb_y,kb_x             wz
        if_z            mov     kb_state,#:storekey
        if_z            jmp     #next_state
                        cmp     kb_y,kb_data          wz      'if match, don't store in buffer
        if_e            shl     kb_y,#16                   ' but do store in high order word
        if_e            wrlong  kb_y,kb_x                     '  of par_break and zero the low
        if_e            mov     kb_state,#:gotbreak              '   order word so test is once only
        if_e            jmp     #next_state              '   order word so test is once only
}
:storekey               rdlong  kb_x,par                   'if room in buffer and key valid, enter
                        sub     kb_x,#1
                        and     kb_x,#$F
                        cmp     kb_x,kb_head         wz
        if_nz           test    kb_data,#$FF       wz
        if_nz           mov     kb_x,par
        if_nz           add     kb_x,#11*4
        if_nz           add     kb_x,kb_head
        if_nz           add     kb_x,kb_head
        if_nz           wrword  kb_data,kb_x
        if_nz           add     kb_head,#1
        if_nz           and     kb_head,#$F

:gotbreak               test    kb_stat,#4         wc      'if not configure flag, done
        if_nc           mov     kb_state,#update           'else configure to update leds
        if_nc           jmp     #next_state             'else configure to update leds
'
'
' Configure keyboard
'
configure               mov     kb_data,#$F3               'set keyboard auto-repeat
                        mov     kb_state_after_transmit,#:auto_repeat_1_transmitted
                        mov     kb_state,#transmit
                        jmp     #next_state
:auto_repeat_1_transmitted
                        mov     kb_data,kb_auto
                        and     kb_data,#%11_11111
                        mov     kb_state_after_transmit,#:auto_repeat_2_transmitted
                        mov     kb_state,#transmit
                        jmp     #next_state
:auto_repeat_2_transmitted
                        mov     kb_data,#$ED               'set keyboard lock-leds
                        mov     kb_state_after_transmit,#:auto_repeat_3_transmitted
                        mov     kb_state,#transmit
                        jmp     #next_state
:auto_repeat_3_transmitted
                        mov     kb_data,kb_locks
                        rev     kb_data,#-3 & $1F
                        test    kb_data,#%100      wc
                        rcl     kb_data,#1
                        and     kb_data,#%111
                        mov     kb_state_after_transmit,#:leds_transmitted
                        mov     kb_state,#transmit
                        jmp     #next_state

:leds_transmitted       mov     kb_x,kb_locks                'insert locks into _states
                        and     kb_x,#%111
                        shl     kb_states+7,#3
                        or      kb_states+7,kb_x
                        ror     kb_states+7,#3

                        mov     kb_present,#1             'set _present

                        mov     kb_state,#update                 'done
                        jmp     #next_state                 'done

{'------------------------------------------------------------------------------------------------------
'
'
' Lookup byte in table
'
look                    ror     kb_data,#2                 'perform lookup
                        movs    :reg,kb_data
                        add     :reg,#table
                        shr     kb_data,#27
                        mov     kb_x,kb_data
:reg                    mov     kb_data,0-0
                        shr     kb_data,kb_x                             
                        and     kb_data,#$FF               'isolate byte
look_ret                ret
}'------------------------------------------------------------------------------------------------------
'
'
' Transmit byte to keyboard
'
transmit                test    kb_pingroup,#1    wz
        if_z            or      dira,kb_cwmask
        if_nz           or      outa,kb_cwmask             'pull clock low

                        rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#13                   'hold clock for ~128us (must be >100us)
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

z1                      test    kb_pingroup,#1    wz
        if_z            or      dira,kb_dwmask
        if_nz           or      outa,kb_dwmask             'pull data low

                        rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#18                  'hold data for ~4us
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

                        test    kb_pingroup,#1    wz
        if_z            xor     dira,kb_cwmask
        if_nz           andn    outa,kb_cwmask             'release clock

                        test    kb_data,#$0FF      wc      'append parity and stop bits to byte
                        muxnc   kb_data,#$100
                        or      kb_data,kb_dlsb

                        mov     kb_x,#10                   'ready 10 bits
transmit_bit

                        'wait until clock low
{wait_c0}               mov     kb_y,kb_tenms                 'set timeout to 10ms

:wloop                  rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#18                  'nap ~4us
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

                        test    kb_crmask,ina      wc      'check required state(s)
                        test    kb_drmask,ina      wz      'loop until got state(s) or timeout
              if_c      djnz    kb_y,#:wloop                '(replaced with c0/c1/c0d0/c1d1)

                        tjz     kb_y,#reset                'if timeout, reset keyboard



                        shr     kb_data,#1         wc      'output data bit
                        test    kb_pingroup,#1    wz
        if_z            muxnc   dira,kb_dwmask
        if_nz           muxnc   outa,kb_dwmask


z3
                        'wait till clock high
{wait_c1}               mov     kb_y,kb_tenms                 'set timeout to 10ms

:wloop                  rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#18                  'nap ~4us
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

                        test    kb_crmask,ina      wc      'check required state(s)
                        test    kb_drmask,ina      wz      'loop until got state(s) or timeout
            if_nc       djnz    kb_y,#:wloop                '(replaced with c0/c1/c0d0/c1d1)

                        tjz     kb_y,#reset                'if timeout, reset keyboard



                        
                        djnz    kb_x,#transmit_bit         'another bit?

' Wait for clock/data 0

wait_c0d0               mov     kb_y,kb_tenms                 'set timeout to 10ms

:wloop                  rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#18                  'nap ~4us
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

                        test    kb_crmask,ina      wc      'check required state(s)
                        test    kb_drmask,ina      wz      'loop until got state(s) or timeout
            if_c_or_nz  djnz    kb_y,#:wloop                '(replaced with c0/c1/c0d0/c1d1)

                        tjz     kb_y,#reset                'if timeout, reset keyboard

' Wait for clock/data 1
wait_c1d1               mov     kb_y,kb_tenms                 'set timeout to 10ms

:wloop                  rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#18                  'nap ~4us
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

                        test    kb_crmask,ina      wc      'check required state(s)
                        test    kb_drmask,ina      wz      'loop until got state(s) or timeout
            if_nc_or_z  djnz    kb_y,#:wloop                '(replaced with c0/c1/c0d0/c1d1)

                        tjz     kb_y,#reset                'if timeout, reset keyboard
'------------------------------------------------------------------------------------------------------

                        mov     kb_state_after_recieve,#:ack_recieved
                        mov     kb_state,#receive_ack       'receive ack byte with timed wait
                        jmp     #next_state
                        
:ack_recieved           cmp     kb_data,#$FA       wz      'if ack error, reset keyboard
        if_nz           mov     kb_state,#reset
        if_nz           jmp     #next_state

                        mov     kb_state,kb_state_after_transmit
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------
'
'
' Receive byte from keyboard
'
receive                 test    kb_crmask,ina wz,wc         'wait indefinitely for initial clock low
            if_nz       jmp     #next_state             ' state already set
'------------------------------------------------------------------------------------------------------                        
receive_ack             mov     kb_x,#11                   'ready 11 bits
receive_bit


{wait_c0}               mov     kb_y,kb_tenms                 'set timeout to 10ms

:wloop                  rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#18                  'nap ~4us
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

                        test    kb_crmask,ina      wc      'check required state(s)
                        test    kb_drmask,ina      wz      'loop until got state(s) or timeout
              if_c      djnz    kb_y,#:wloop                '(replaced with c0/c1/c0d0/c1d1)

                        tjz     kb_y,#reset                'if timeout, reset keyboard

z2
                        
                        rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#16                  'pause ~16us
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state

'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

                        test    kb_drmask,ina      wc      'input data bit
                        rcr     kb_data,#1
z4


                        'call    #wait_c1
{wait_c1}               mov     kb_y,kb_tenms                 'set timeout to 10ms

:wloop                  rdlong  kb_t,#0                    'get clkfreq
                        shr     kb_t,#18                  'nap ~4us
                        min     kb_t,#3                    'ensure waitcnt won't snag
                        add     kb_t,cnt                   'add cnt to time
                        mov     kb_state,#:nap_cnt_loop
                        jmp     #next_state
'------------------------------------------------------------------------------------------------------                        
:nap_cnt_loop           cmp     kb_t,cnt wz,wc
            if_a        mov     kb_state,#:nap_cnt_loop
            if_a        jmp     #next_state

                        test    kb_crmask,ina      wc      'check required state(s)
                        test    kb_drmask,ina      wz      'loop until got state(s) or timeout
            if_nc       djnz    kb_y,#:wloop                '(replaced with c0/c1/c0d0/c1d1)

                        tjz     kb_y,#reset                'if timeout, reset keyboard

                        djnz    kb_x,#receive_bit          'another bit?

                        shr     kb_data,#22                'align byte
                        test    kb_data,#$1FF      wc      'if parity error, reset keyboard
        if_nc           jmp     #reset
rand                    and     kb_data,#$FF               'isolate byte

                        mov     kb_state,kb_state_after_recieve
                        jmp     #next_state





































' /////////////////////////////////////////////////////////////////////////////
' dividefract:
' Perform 2^32 * r1/r2, result stored in r3 (useful for TV calc)
' This is taken from the tv driver.
' NOTE: It divides a bottom heavy fraction e.g. 1/2 and gives the result as a 32-bit fraction.
' /////////////////////////////////////////////////////////////////////////////
dividefract
                        mov     r0,#32+1
:loop                   cmpsub  r1,r2           wc
                        rcl     r3,#1
                        shl     r1,#1
                        djnz    r0,#:loop
dividefract_ret         ret                             '+140




' Initialized data
inc_dest                long    1 << 9 << 0

' Video (TV) Registers
tvport_mask             long    %0000_0111<<24
v_freq                  long    0
v_clkfreq               long    0


'Pixel streams
'  These are shifted out of the VSU to the right, so lowest bits are actioned first.
'
hsync                   long    %%11_0000_1_2222222_11  ' Used with NTSC_control_signal_palette so:
                                                        '      0 = blanking level
                                                        '      1 = Black
                                                        '      2 = Color NTSC_control_signal_palette (yellow at zero value)
vsync_high_1            long    %%11111111111_222_11
vsync_high_2            long    %%1111111111111111
vsync_low_1             long    %%22222222222222_11
vsync_low_2             long    %%1_222222222222222
border                  long    %%0000000000000000

' Some unimportant irrelevant constants for generating demo user display etc.
line_loop               long    0
tile_loop               long    0
vsync_flag              long    $00000001

'Palettes
'  These are always 4 colors (or blanking level) stored in reverse order:
'                               Color3_Color2_Color1_Color0
'
black_border_in_color0  long    $00_00_00_02
brown_border_in_color0  long    $00_00_00_02

' Keyboard initialised data
kb_dlsb                 long    1 << 9
kb_tenms                long    10_000 / 4
'
'
' Lookup table
'                               ascii   scan    extkey  regkey  ()=keypad
'
{table                   word    $0000   '00
                        word    $00D8   '01             F9
                        word    $0000   '02
                        word    $00D4   '03             F5
                        word    $00D2   '04             F3
                        word    $00D0   '05             F1
                        word    $00D1   '06             F2
                        word    $00DB   '07             F12
                        word    $0000   '08
                        word    $00D9   '09             F10
                        word    $00D7   '0A             F8
                        word    $00D5   '0B             F6
                        word    $00D3   '0C             F4
                        word    $0009   '0D             Tab
                        word    $0060   '0E             `
                        word    $0000   '0F
                        word    $0000   '10
                        word    $F5F4   '11     Alt-R   Alt-L
                        word    $00F0   '12             Shift-L
                        word    $0000   '13
                        word    $F3F2   '14     Ctrl-R  Ctrl-L
                        word    $0071   '15             q
                        word    $0031   '16             1
                        word    $0000   '17
                        word    $0000   '18
                        word    $0000   '19
                        word    $007A   '1A             z
                        word    $0073   '1B             s
                        word    $0061   '1C             a
                        word    $0077   '1D             w
                        word    $0032   '1E             2
                        word    $F600   '1F     Win-L
                        word    $0000   '20
                        word    $0063   '21             c
                        word    $0078   '22             x
                        word    $0064   '23             d
                        word    $0065   '24             e
                        word    $0034   '25             4
                        word    $0033   '26             3
                        word    $F700   '27     Win-R
                        word    $0000   '28
                        word    $0020   '29             Space
                        word    $0076   '2A             v
                        word    $0066   '2B             f
                        word    $0074   '2C             t
                        word    $0072   '2D             r
                        word    $0035   '2E             5
                        word    $CC00   '2F     Apps
                        word    $0000   '30
                        word    $006E   '31             n
                        word    $0062   '32             b
                        word    $0068   '33             h
                        word    $0067   '34             g
                        word    $0079   '35             y
                        word    $0036   '36             6
                        word    $CD00   '37     Power
                        word    $0000   '38
                        word    $0000   '39
                        word    $006D   '3A             m
                        word    $006A   '3B             j
                        word    $0075   '3C             u
                        word    $0037   '3D             7
                        word    $0038   '3E             8
                        word    $CE00   '3F     Sleep
                        word    $0000   '40
                        word    $002C   '41             ,
                        word    $006B   '42             k
                        word    $0069   '43             i
                        word    $006F   '44             o
                        word    $0030   '45             0
                        word    $0039   '46             9
                        word    $0000   '47
                        word    $0000   '48
                        word    $002E   '49             .
                        word    $EF2F   '4A     (/)     /
                        word    $006C   '4B             l
                        word    $003B   '4C             ;
                        word    $0070   '4D             p
                        word    $002D   '4E             -
                        word    $0000   '4F
                        word    $0000   '50
                        word    $0000   '51
                        word    $0027   '52             '
                        word    $0000   '53
                        word    $005B   '54             [
                        word    $003D   '55             =
                        word    $0000   '56
                        word    $0000   '57
                        word    $00DE   '58             CapsLock
                        word    $00F1   '59             Shift-R
                        word    $EB0D   '5A     (Enter) Enter
                        word    $005D   '5B             ]
                        word    $0000   '5C
                        word    $005C   '5D             \
                        word    $CF00   '5E     WakeUp
                        word    $0000   '5F
                        word    $0000   '60
                        word    $0000   '61
                        word    $0000   '62
                        word    $0000   '63
                        word    $0000   '64
                        word    $0000   '65
                        word    $00C8   '66             BackSpace
                        word    $0000   '67
                        word    $0000   '68
                        word    $C5E1   '69     End     (1)
                        word    $0000   '6A
                        word    $C0E4   '6B     Left    (4)
                        word    $C4E7   '6C     Home    (7)
                        word    $0000   '6D
                        word    $0000   '6E
                        word    $0000   '6F
                        word    $CAE0   '70     Insert  (0)
                        word    $C9EA   '71     Delete  (.)
                        word    $C3E2   '72     Down    (2)
                        word    $00E5   '73             (5)
                        word    $C1E6   '74     Right   (6)
                        word    $C2E8   '75     Up      (8)
                        word    $00CB   '76             Esc
                        word    $00DF   '77             NumLock
                        word    $00DA   '78             F11
                        word    $00EC   '79             (+)
                        word    $C7E3   '7A     PageDn  (3)
                        word    $00ED   '7B             (-)
                        word    $DCEE   '7C     PrScr   (*)
                        word    $C6E9   '7D     PageUp  (9)
                        word    $00DD   '7E             ScrLock
                        word    $0000   '7F
                        word    $0000   '80
                        word    $0000   '81
                        word    $0000   '82
                        word    $00D6   '83             F7
}
keypad1                 byte    $CA, $C5, $C3, $C7, $C0, 0, $C1, $C4, $C2, $C6, $C9, $0D, "+-*/"

keypad2                 byte    "0123456789.", $0D, "+-*/"

shift1                  byte    "{|}", 0, 0, "~"

shift2                  byte    $22, 0, 0, 0, 0, "<_>?)!@#$%^&*(", 0, ":", 0, "+"


pixels                  res     1
ptr                     res     1
current_line            res     1

' General Purpose Registers
r0                      res     1
r1                      res     1
r2                      res     1
r3                      res     1


params                             
tv_scanline_request_adr res     1
tv_scanline_buffer_adr  res     1
kb_tail                 res     1   'key buffer tail        read/write      (20 contiguous longs)
kb_head                 res     1   'key buffer head        read-only
kb_present              res     1   'keyboard present       read-only
kb_states               res     8   'key states (256 bits)  read-only
kb_keys                 res     8   'key buffer (16 words)  read-only       (also used to pass initial parameters)
kb_break                res     1   'break key code, flag   read/write
kb_pingroup             res     1
kb_locks                res     1
kb_auto                 res     1
'end of params

kb_state                   res     1
kb_state_after_transmit    res     1
kb_state_after_recieve     res     1
kb_state_after_look        res     1

kb_dwmask               res     1       'output driver mask for data
kb_drmask               res     1       'input mask for data
kb_cwmask               res     1       'output driver mask for clock
kb_crmask               res     1       'input mask for data
kb_stat                 res     1
kb_data                 res     1
kb_x                    res     1
kb_y                    res     1
kb_t                    res     1
      